// The Screen class is a derivative of Window that writes the resulting
// data onto the physical screen

#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <sys/socket.h> // Some libc/linux versions seem to have iovec there :(
#include "mcl.h"
#include "cui.h"
#include "Buffer.h"
#include "Curses.h"
#include "Session.h"

// This isn't too pretty
int get_screen_size_x()
{
	// Screen dimensions
	struct winsize size;
	if (ioctl(STDIN_FILENO, TIOCGWINSZ, (char *) &size) < 0)
		error ("get_screen_size:ioctl:%m");

	return size.ws_col;		
}

int get_screen_size_y()
{
	// Screen dimensions
	struct winsize size;
	if (ioctl(STDIN_FILENO, TIOCGWINSZ, (char *) &size) < 0)
		error ("get_screen_size:ioctl:%m");

	return size.ws_row;		
}
		
		

// Initialization. Figure out what dimensions we have here 
Screen::Screen() : Window(NULL, get_screen_size_x(), get_screen_size_y()), fd(0) {
    // Open our fd
    char *tty = ttyname(STDIN_FILENO);
    int ttyno;
    char buf[256];
    usingVirtual = true;

    // pre 2.4
    if (1 == (sscanf(tty, "/dev/tty%d", &ttyno)))
        sprintf (buf, "/dev/vcsa%d", ttyno);
    // 2.4 with devs (noted by moon@deathmoon.com
    else if (1 == (sscanf(tty, "/dev/vc/%d", &ttyno)))
        sprintf (buf, "/dev/vcc/a%d", ttyno);
    else {
        usingVirtual = false;
        scr_x = scr_y = scr_w = scr_h = 0;
        last_screen = new attrib[width * height];
        memcpy(last_screen, canvas, width * height * sizeof(attrib));
        write(STDOUT_FILENO, CLEAR_SCREEN, strlen(CLEAR_SCREEN));
        out = new Buffer(32000);
    }

    if (usingVirtual) {
      if ((fd = open (buf, O_WRONLY)) < 0) {
	fprintf (stderr, "\nFailed to open %s: %m. \nPerhaps your permissions are wrong?\n\n", buf);
	exit (EXIT_FAILURE);
      }
    }

    init_curses(usingVirtual);
}

// Refresh the screen
bool Screen::refreshVirtual() {
    unsigned char cursor_pos[2];
    
    struct iovec write_struct[2] =
    {
        {(void*) &cursor_pos,	2},
        {(void*) canvas,		width*height*sizeof(*canvas)}
    };
    
    if (dirty)
        clear();
    
    if (Window::wrefresh())
    {
        
        lseek(fd,2,SEEK_SET);	// Go to beginning of the vcsa, but not the first
        // 2 bytes which are selection pos or somethign
        
        // Write cursor position, then canvas data to the device
        
        cursor_pos[0] = cursor_x;
        cursor_pos[1] = cursor_y;
        
        while (writev (fd, write_struct, 2) < 0)
            if (errno != EINTR && errno != EAGAIN)
                error ("Screen::srefresh():writev");
        
        return true;
    }
    else
        return false;
}

bool Screen::srefresh() {
    if (usingVirtual)
        return refreshVirtual();
    else
        return refreshTTY();
}

// Why this?
// ANSI colors and the colors used by /dev/vcsa do not fit together
// Internally we use the /dev/vcsa colors so we must translate here
static int reverse_color_conv_table[8] =  {
    0,4,2,6,1,5,3,7
};

// 8-bit CSI - use this if your terminal supports it!
// #define CSI "\x9B"
// 7-bit CSI
#define CSI "\e["

// TODO: split it up so we can change fg and bg seperately
const char* Screen::getColorCode (int color, bool setBackground) const {
    int fg_color = 30 + reverse_color_conv_table[color & 0x07];
    int bold = color & fg_bold;
    int bg_color = 40 + reverse_color_conv_table[color >> 4];
    static char buf[64];
    
    // Xterm doesn't like fg_white|bg_black ?!
    if (fg_color == 37 &&  bg_color == 40 && !bold)
        sprintf(buf, CSI "0m");
    else {
        char bg[8];
        if (setBackground)
            sprintf(bg, "%d;", bg_color);
        else
            bg[0] = NUL;
        
        if (bold)
            sprintf(buf, CSI "1;%s%dm", bg, fg_color);
        else
            sprintf(buf, CSI "0;%s%dm", bg, fg_color);
    }

    return buf;
}

void Screen::setTTYColor(int color, bool setBackground) {
    const char *code = getColorCode(color, setBackground);
    globalStats.ctrl_chars += out->strcat(code);
}

// Print a single character. If it's a graphical character, and ACS is not yet enabled it, enable it
void Screen::printCharacter(int c, bool& acs_enabled) {
    if (c >= SC_BASE && c < SC_END && smacs) {
        if (!acs_enabled)
            out->strcat(smacs);
        out->chrcat(real_special_chars[c-SC_BASE]);
        acs_enabled = true;
    } else {
        if (acs_enabled && rmacs) {
            out->strcat(rmacs);
            acs_enabled = false;
        }
        out->chrcat(c);
    }
}


// This is Y,X, both starting at 1
#define VT_GOTO CSI "%d;%dH"

// goes to 1,1
#define VT_HOME CSI "H" 
#define MAX_SCROLL_TRY 5

// some more optimization could be useful here!
// Checksum rows? (could error)
// do a memcp before running through it?
// probably not worth it, it's the output that's most important
bool Screen::refreshTTY() {
    
	if (!Window::wrefresh()) return FALSE;
	for (int i = 0; i < width*height - 1; i++) {
		if ((i % width) == 0) move (i / width, 0);
		addch(canvas[i] & 0xff);
	}
	move(cursor_y, cursor_x);
	refresh();
	return true;
}

Screen::~Screen() {
    if (usingVirtual)
        close (fd);
    else endwin();
}

void Screen::set_cursor(int _x, int _y) {
	cursor_x = _x;
	cursor_y = _y;
}

void Screen::setScrollingRegion(int x, int y, int w, int h) {
    scr_x = x; scr_y = y; scr_w = w; scr_h = h;
}

void Screen::flash() {
	if (config->getOption(opt_beep))
		write (STDIN_FILENO, "\a",1);
}

bool Screen::keypress(int key) {
    // Check for alt-X
    if (key >= key_alt_1 && key < key_alt_9)  {
        Window *w = Numbered::find(key-key_alt_1+1);
        if (w)
        {
            w->show(true);
            w->popUp();
            return true;
        }
    }
    else if (key == key_ctrl_l) { // redraw
        if (!usingVirtual)
            memset(last_screen, 0, width*height*sizeof(attrib));
        return true;
    }
    
    if (!Window::keypress(key))
        if (!currentSession || !currentSession->expand_macros(key))
            status->setf ("Keycode %d (%s) was not handled by any object", key, key_name(key));

    return true;
}
